# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re
import math

try:
    import libsbml
except ImportError as err:
    raise ImportError('libsbml is required to convert SBML files for GillesPy.') from err

import numpy as np

import gillespy2
from gillespy2.core.gillespyError import InvalidModelError, SBMLError
from gillespy2.core import log

PYTHON_RESERVED_WORDS = [
    'False', 'def', 'if', 'raise',
    'None', 'del', 'import', 'return',
    'True', 'elif', 'in', 'try',
    'and', 'else', 'is', 'while',
    'as', 'except', 'lambda', 'with',
    'assert', 'finally', 'nonlocal', 'yield',
    'break', 'for', 'not',
    'class', 'form', 'or',
    'continue', 'global', 'pass',
]
PREFIX_INVALID_SBML_NAME = 'GPY2__' #'lambda' -> 'k_lambda'


init_state = {'INF': np.inf, 'NaN': np.nan}
postponed_evals = {}
eval_globals = math.__dict__.copy()
def piecewise(*args):
    args = list(args)
    sol = None
    if len(args) % 2:
        args.append(True)
    for i, arg in enumerate(args):
        if not i % 2:
            continue
        if arg:
            sol = args[i-1]
            break
    return sol
eval_globals['piecewise'] = piecewise

def __read_sbml_model(filename):

    document = libsbml.readSBML(filename)
    errors = []
    error_count = document.getNumErrors()
    if error_count > 0:
        for i in range(error_count):
            error = document.getError(i)
            converter_code = 0
            converter_code = -10

            errmsg = f"SBML {error.getSeverityAsString()}, code {error.getErrorId()}, "
            errmsg += f"line {error.getLine()}: {error.getMessage()}"
            errors.append([errmsg, converter_code])
    if min([code for error, code in errors] + [0]) < 0:
        return None, errors
    sbml_model = document.getModel()

    return sbml_model, errors

def __get_math(formula):
    math_str = libsbml.formulaToL3String(formula)
    replacements = {
        r'\bln\b': 'log',
        r'\^': '**',
        r'\&\&': 'and',
        r'\|\|': 'or',
        }
    for old, new in replacements.items():
        math_str = re.sub(old, new, math_str)
    for rword in PYTHON_RESERVED_WORDS:
        reg = r'\b' + rword + r'\b'
        if re.search(reg, math_str):
            sub = PREFIX_INVALID_SBML_NAME + rword
            log.warning("'%s' is an invalid name in GillesPy2, changing to '%s'",rword,sub)
            math_str = re.sub(reg, sub, math_str)
    return math_str

def __get_species(sbml_model, gillespy_model, errors):

    for i in range(sbml_model.getNumSpecies()):
        species = sbml_model.getSpecies(i)
        name = species.getId()
        population = None
        concentration = None
        value = None
        mode = 'continuous'
        cid = sbml_model.getCompartment(species.getCompartment()).getName()
        if species.isSetInitialAmount(): # Amount is provided
            population = species.getInitialAmount()
        if species.isSetInitialConcentration(): # Concentration is provided
            concentration = species.getInitialConcentration()
        if population is None and concentration is None: # Assignment
            value = 0
        elif species.getHasOnlySubstanceUnits(): # Treat as population
            if population is not None: # If population is provided
                value = population
                int_value = int(species.getInitialAmount())
                if population == int_value:
                    population = int_value
            else: # Else convert concentration to population
                postponed_evals[name] = f'{name} * {cid}'
                value = concentration
        else: # Treat as concentration
            if concentration is not None: # If concentration is provided
                value = concentration
            else: # Else convert population to concentration
                postponed_evals[name] = f'{name} / {cid}'
                value = population

        constant = species.getConstant()
        boundary_condition = species.getBoundaryCondition()
        is_negative = value < 0
        if name in PYTHON_RESERVED_WORDS:
            name =  PREFIX_INVALID_SBML_NAME + name
        gillespy_species = gillespy2.Species(name=name, initial_value=value,
                                                allow_negative_populations=is_negative, mode=mode,
                                                constant=constant, boundary_condition=boundary_condition)
        gillespy_model.add_species([gillespy_species])
        init_state[name] = value

def __get_parameters(sbml_model, gillespy_model):

    for i in range(sbml_model.getNumParameters()):
        parameter = sbml_model.getParameter(i)
        name = parameter.getId()
        if parameter.isSetValue():
            value = parameter.getValue()
        else:
            value = 0
        init_state[name] = value

        # GillesPy2 represents non-constant parameters as species
        if name in PYTHON_RESERVED_WORDS:
            name =  PREFIX_INVALID_SBML_NAME + name
        if parameter.isSetConstant():
            gillespy_parameter = gillespy2.Parameter(name=name, expression=value)
            gillespy_model.add_parameter([gillespy_parameter])
        else:
            gillespy_species = gillespy2.Species(name=name,initial_value=value)
            gillespy_model.add_species([gillespy_species])

def __get_compartments(sbml_model, gillespy_model):
    for i in range(sbml_model.getNumCompartments()):
        compartment = sbml_model.getCompartment(i)
        name = compartment.getId()
        if compartment.isSetSize():
            value = compartment.getSize()
        else:
            value = 1

        if name == "vol":
            gillespy_model.volume = value
        else:
            if name in PYTHON_RESERVED_WORDS:
                name =  PREFIX_INVALID_SBML_NAME + name
            gillespy_parameter = gillespy2.Parameter(name=name, expression=value)
            init_state[name] = value
            gillespy_model.add_parameter([gillespy_parameter])

def __traverse_math(node, old_id, new_id):
    if node is None:
        return
    for i in range(node.getNumChildren()):
        if node.getChild(i).getName() == old_id:
            new_node = libsbml.ASTNode()
            new_node.setName(new_id)
            node.replaceChild(i, new_node)
        __traverse_math(node.getChild(i), old_id, new_id)

def __get_kinetic_law(sbml_model, gillespy_model, reaction):
    kinetic_law = reaction.getKineticLaw()

    if kinetic_law is None:
        raise InvalidModelError(
            f"Failed to load SBML model: Reaction '{reaction}' is missing its propensity function."
        )

    tree = kinetic_law.getMath()
    params = kinetic_law.getListOfParameters()
    local_params = kinetic_law.getListOfLocalParameters()
    for i in range(kinetic_law.getNumLocalParameters()):
        local_param = local_params.get(i)
        old_id = local_param.getId()
        new_id = (f'{reaction.getId()}_{local_param.getId()}')
        __traverse_math(tree, old_id, new_id)
        local_param.setId(new_id)
        name = new_id
        if name in PYTHON_RESERVED_WORDS:
            name =  PREFIX_INVALID_SBML_NAME + name
        gillespy_parameter = gillespy2.Parameter(name=name, expression=local_param.getValue())
        gillespy_model.add_parameter([gillespy_parameter])
    for i in range(kinetic_law.getNumParameters()):
        param = params.get(i)
        if not param.getId() in gillespy_model.listOfParameters:
            name = param.getId()
            if name in PYTHON_RESERVED_WORDS:
                name =  PREFIX_INVALID_SBML_NAME + name
            gillespy_parameter = gillespy2.Parameter(name=name, expression=param.getValue())
            gillespy_model.add_parameter([gillespy_parameter])
    return tree

def __get_reactions(sbml_model, gillespy_model, errors):
    # reactions
    for i in range(sbml_model.getNumReactions()):
        reaction = sbml_model.getReaction(i)
        name = reaction.getId()
        tree = __get_kinetic_law(sbml_model, gillespy_model, reaction)
        propensity = __get_math(tree)
        reactants = {}
        products = {}

        r_set = set()
        p_set = set()
        # get reactants
        for j in range(reaction.getNumReactants()):
            reactant = reaction.getReactant(j)
            species = reactant.getSpecies()

            if species == "EmptySet":
                continue
            stoichiometry = reactant.getStoichiometry()
            if isinstance(stoichiometry, float):
                if int(stoichiometry) != stoichiometry:
                    logmsg = f"Reaction {name} contains a float stoichiometry for reactant {species}.  "
                    logmsg += "Please check your model as this may cause inaccuracies in the results."
                    gillespy2.log.warning(logmsg)
                stoichiometry = int(stoichiometry)

            if species in r_set:
                reactants[species] += stoichiometry
            else:
                r_set.add(species)
                reactants[species] = stoichiometry

        # get products
        for j in range(reaction.getNumProducts()):
            product = reaction.getProduct(j)
            species = product.getSpecies()

            if species == "EmptySet":
                continue
            stoichiometry = product.getStoichiometry()
            if isinstance(stoichiometry, float):
                if int(stoichiometry) != stoichiometry:
                    logmsg = f"Reaction {name} contains a float stoichiometry for product {species}.  "
                    logmsg += "Please check your model as this may cause inaccuracies in the results."
                    gillespy2.log.warning(logmsg)
                stoichiometry = int(stoichiometry)

            if species in p_set:
                products[species] += stoichiometry
            else:
                p_set.add(species)
                products[species] = stoichiometry

        gillespy_reaction = gillespy2.Reaction(name=name, reactants=reactants, products=products,
                                             propensity_function=propensity)

        gillespy_model.add_reaction([gillespy_reaction])

def __get_rules(sbml_model, gillespy_model, errors):
    for i in range(sbml_model.getNumRules()):
        rule = sbml_model.getRule(i)

        # If the SBML object does not contain an ID attribute create a unique rule_name from the variable name.
        rule_name = rule.getIdAttribute()
        if rule_name == "":
            rule_name = f"ar_{rule.getId()}"

        rule_variable = rule.getVariable()
        rule_string = __get_math(rule.getMath())
        if rule_variable in gillespy_model.listOfParameters:
            # Treat Non-Constant Parameters as Species
            value = gillespy_model.listOfParameters[rule_variable].expression
            name = rule_variable
            if name in PYTHON_RESERVED_WORDS:
                name =  PREFIX_INVALID_SBML_NAME + name
            species = gillespy2.Species(name=name,
                                        initial_value=value,
                                        allow_negative_populations=True,
                                        mode='continuous')
            gillespy_model.delete_parameter(rule_variable)
            gillespy_model.add_species([species])

        t = []

        if rule.isAssignment():
            _ = eval(rule_string, {**eval_globals, **init_state})
            postponed_evals[rule_variable] = rule_string
            gillespy_rule = gillespy2.AssignmentRule(name=rule_name, variable=rule_variable,
                formula=rule_string)
            gillespy_model.add_assignment_rule(gillespy_rule)
            init_state[gillespy_rule.variable.name]=eval(gillespy_rule.formula, {**init_state, **eval_globals})

        if rule.isRate():
            gillespy_rule = gillespy2.RateRule(name=rule_name, variable=rule_variable,
                formula=rule_string)
            gillespy_model.add_rate_rule(gillespy_rule)

        if rule.isAlgebraic():
            t.append('algebraic')

            if len(t) > 0:
                t[0] = t[0].capitalize()

                msg = ", ".join(t)
                msg += " rule"
            else:
                msg = "Rule"

            errmsg = f"{msg} '{rule.getId()}' found on line '{rule.getLine()}' with equation "
            errmsg += f"'{libsbml.formulaToString(rule.getMath())}'. gillespy does not support SBML Algebraic Rules."
            errors.append([errmsg, -5])

def __get_constraints(sbml_model, gillespy_model, errors):
    for i in range(sbml_model.getNumConstraints()):
        constraint = sbml_model.getConstraint(i)

        errmsg = f"Constraint '{constraint.getId()}' found on line '{constraint.getLine()}' with equation "
        errmsg += f"'{libsbml.formulaToString(constraint.getMath())}'. gillespy does not support SBML Constraints"
        errors.append([errmsg, -5])

def __get_function_definitions(sbml_model, gillespy_model):
    # TODO:
    # DOES NOT CURRENTLY SUPPORT ALL MATHML
    # ALSO DOES NOT SUPPORT NON-MATHML
    for i in range(sbml_model.getNumFunctionDefinitions()):
        function = sbml_model.getFunctionDefinition(i)
        function_name = function.getId()
        function_tree = function.getMath()
        num_nodes = function_tree.getNumChildren()
        function_args = [function_tree.getChild(i).getName() for i in range(num_nodes-1)]
        function_string = __get_math(function_tree.getChild(num_nodes-1))
        gillespy_function = gillespy2.FunctionDefinition(
            name=function_name, function=function_string, args=function_args
        )
        gillespy_model.add_function_definition(gillespy_function)
        init_state[gillespy_function.name] = gillespy_function.function_string

def __get_events(sbml_model, gillespy_model):
    for i in range(sbml_model.getNumEvents()):
        event = sbml_model.getEvent(i)
        gillespy_assignments = []

        trigger = event.getTrigger()
        delay = event.getDelay()
        if delay is not None:
            delay = libsbml.formulaToL3String(delay.getMath())
        expression=libsbml.formulaToL3String(trigger.getMath())
        expression = expression.replace('&&', ' and ').replace('||', ' or ')
        initial_value = trigger.getInitialValue()
        persistent = trigger.getPersistent()
        use_values_from_trigger_time = event.getUseValuesFromTriggerTime()
        gillespy_trigger = gillespy2.EventTrigger(
            expression=expression, initial_value=initial_value, persistent=persistent
        )
        assignments = event.getListOfEventAssignments()
        for assign in assignments:
            # Convert Non-Constant Parameter to Species
            if assign.getVariable() in gillespy_model.listOfParameters:
                name = assign.getVariable()
                if name in PYTHON_RESERVED_WORDS:
                    name =  PREFIX_INVALID_SBML_NAME + name
                gillespy_species = gillespy2.Species(
                    name=name,
                    initial_value=gillespy_model.listOfParameters[assign.getVariable()].expression,
                    mode='continuous', allow_negative_populations=True
                )
                gillespy_model.delete_parameter(assign.getVariable())
                gillespy_model.add_species([gillespy_species])

            gillespy_assignment = gillespy2.EventAssignment(variable=assign.getVariable(),
                expression=__get_math(assign.getMath()))
            gillespy_assignments.append(gillespy_assignment)
        gillespy_event = gillespy2.Event(
            name=event.getId(), trigger=gillespy_trigger,
            assignments=gillespy_assignments, delay=delay,
            use_values_from_trigger_time=use_values_from_trigger_time)
        gillespy_model.add_event(gillespy_event)

def __get_initial_assignments(sbml_model, gillespy_model):

    for i in range(sbml_model.getNumInitialAssignments()):
        init_assign = sbml_model.getInitialAssignment(i)
        variable = init_assign.getId()
        expression = __get_math(init_assign.getMath())
        assigned_value = eval(expression, {**init_state, **eval_globals})
        init_state[variable] = assigned_value
        if assigned_value != assigned_value:
            assigned_value = expression
            postponed_evals[variable] = expression

        if variable in gillespy_model.listOfSpecies:
            gillespy_model.listOfSpecies[variable].initial_value = assigned_value
        elif variable in gillespy_model.listOfParameters:
            gillespy_model.listOfParameters[variable].expression = str(assigned_value)

def __resolve_evals(gillespy_model, init_state):
    while True:
        successful = []
        if len(postponed_evals):
            for var, expr in postponed_evals.items():
                try:
                    assigned_value = eval(expr, {**eval_globals, **init_state})
                except Exception:
                    assigned_value = np.nan
                if assigned_value == assigned_value:
                    successful.append(var)
                    init_state[var] = assigned_value
                    if var in gillespy_model.listOfSpecies:
                        gillespy_model.listOfSpecies[var].initial_value = assigned_value
                    elif var in gillespy_model.listOfParameters:
                        gillespy_model.listOfParameters[var].value = assigned_value
        if len(successful) <= 0:
            break
        for var in successful:
            del postponed_evals[var]

def convert(filename, model_name=None, gillespy_model=None, report_silently_with_sbml_error=False):


    sbml_model, errors = __read_sbml_model(filename)

    if sbml_model is None:
        if report_silently_with_sbml_error:
            return None, errors
        errs = '\n\t'.join(errors)
        raise SBMLError(f"SBML model import failed.  Reason Given: \n\t{errs}")

    if len(errors) > 0 and not report_silently_with_sbml_error:
        from gillespy2 import log # pylint: disable=import-outside-toplevel
        errs = '\n\t'.join(errors)
        errmsg = f"Error were detected in the SBML model.  Error: \n\t{errs}"
        log.warning(errmsg)

    if model_name is None:
        model_name = sbml_model.getName()
    if gillespy_model is None:
        gillespy_model = gillespy2.Model(name=model_name)
    gillespy_model.units = "concentration"

    __get_function_definitions(sbml_model, gillespy_model)
    __get_parameters(sbml_model, gillespy_model)
    __get_species(sbml_model, gillespy_model, errors)
    __get_compartments(sbml_model, gillespy_model)
    __get_reactions(sbml_model, gillespy_model, errors)
    __get_rules(sbml_model, gillespy_model, errors)
    __get_constraints(sbml_model, gillespy_model, errors)
    __get_events(sbml_model, gillespy_model)
    __get_initial_assignments(sbml_model, gillespy_model)
    __resolve_evals(gillespy_model, init_state)

    return gillespy_model, errors



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np

from gillespy2.solvers.cpp.c_decoder import IterativeSimDecoder
from gillespy2.solvers.utilities import solverutils as cutils
from gillespy2.core import GillesPySolver, Model
from gillespy2.core.gillespyError import *
from gillespy2.core import Results

from .c_solver import CSolver, SimulationReturnCode

class ODECSolver(GillesPySolver, CSolver):
    """
    This Solver produces the deterministic continuous solution via Ordinary Differential Equations.
    Uses integrators from SUNDIALS to perform calculations used to produce solutions.
    """

    name = "ODECSolver"
    target = "ode"

    @staticmethod
    def get_supported_integrator_options():
        return {
            "rtol",
            "atol",
            "max_step",
        }

    @classmethod
    def get_solver_settings(cls):
        """
        Returns a list of arguments supported by odc_c_solver.run.
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        :rtype: tuple
        """
        return ('model', 't', 'number_of_trajectories', 'timeout', 'increment', 'seed', 'debug', 'profile')

    def run(self=None, model: Model = None, t: int = None, number_of_trajectories: int = 1, timeout: int = 0,
            increment: int = None, seed: int = None, debug: bool = False, profile: bool = False, variables={},
            resume=None, live_output: str = None, live_output_options: dict = {}, integrator_options: "dict[str, float]" = None, **kwargs):

        """
        :param model: The model on which the solver will operate. (Deprecated)
        :type model: gillespy2.Model

        :param t: End time of simulation.
        :type t: int

        :param number_of_trajectories: Number of trajectories to simulate.
            This is deterministic and will always have same results.
        :type number_of_trajectories: int

        :param timeout: If set, if simulation takes longer than timeout, will exit.
        :type timeout: int

        :param increment: Time step increment for plotting.
        :type increment: float

        :param seed: The random seed for the simulation. Optional, defaults to None.
        :type seed: int

        :param variables: Dictionary of species and their data that will override existing species data.
        :type variables: dict

        :param resume: Result of a previously run simulation, to be resumed.
        :type resume: gillespy2.Results

        :param live_output: The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :type live_output: str

        :param live_output_options: Dictionary that contains options for live_output. By default {"interval":1}.
            "interval" specifies seconds between displaying.
            "clear_output" specifies if display should be refreshed with each display.
        :type live_output_options:  dict

        :returns: A result object containing the results of the simulation.
        :rtype: gillespy2.Results
        """

        from gillespy2 import log

        if self is None:
            # Post deprecation block
            # raise SimulationError("ODECSolver must be instantiated to run the simulation")
            # Pre deprecation block
            log.warning(
                """
                `gillespy2.Model.run(solver=ODECSolver)` is deprecated.

                You should use `gillespy2.Model.run(solver=ODECSolver(model=gillespy2.Model))
                Future releases of GillesPy2 may not support this feature.
                """
            )
            self = ODECSolver(model, resume=resume)

        if model is not None:
            log.warning('model = gillespy2.model is deprecated. Future releases '
                        'of GillesPy2 may not support this feature.')
        if self.model is None:
            if model is None:
                raise SimulationError("A model is required to run the simulation.")
            self._set_model(model=model)

        self.model.compile_prep()
        self.validate_model(self.model, model)
        self.validate_sbml_features(model=self.model)

        self.validate_tspan(increment=increment, t=t)
        if increment is None:
            increment = self.model.tspan[-1] - self.model.tspan[-2]
        if t is None:
            t = self.model.tspan[-1]

        # Validate parameters prior to running the model.
        self._validate_type(variables, dict, "'variables' argument must be a dictionary.")

        self._validate_resume(t, resume)
        self._validate_kwargs(**kwargs)

        if resume is not None:
            t = abs(t - int(resume["time"][-1]))

        number_timesteps = int(round(t / increment + 1))

        args = {
            "timesteps": number_timesteps,
            "end": t,
            "increment": increment,
            "interval": str(number_timesteps),
        }

        if self.variable:
            populations = cutils.update_species_init_values(self.model.listOfSpecies, self.species, variables, resume)
            parameter_values = cutils.change_param_values(self.model.listOfParameters, self.parameters, self.model.volume, variables)

            args.update({
                "init_pop": populations,
                "parameters": parameter_values
            })
        elif variables:
            log.warning("'variables' argument ignored, because solver has variable=False.")

        if integrator_options is not None:
            integrator_options = ODECSolver.validate_integrator_options(integrator_options)
            args.update(integrator_options)

        seed = self._validate_seed(seed)
        if seed is not None:
            args.update({
                "seed": seed
            })

        if live_output is not None:
            live_output_options['type'] = live_output
            display_args = {
                "model": self.model, "number_of_trajectories": number_of_trajectories, "timeline": np.linspace(0, t, number_timesteps),
                "live_output_options": live_output_options, "resume": bool(resume)
            }
        else:
            display_args = None

        args = self._make_args(args)
        decoder = IterativeSimDecoder.create_default(1, number_timesteps, len(self.model.listOfSpecies))

        sim_exec = self._build(self.model, self.target, self.variable, False)
        sim_status = self._run(sim_exec, args, decoder, timeout, display_args)

        if sim_status == SimulationReturnCode.FAILED:
            raise ExecutionError("Error encountered while running simulation C++ file:\n"
                f"Return code: {int(sim_status)}.\n")

        trajectories, time_stopped = decoder.get_output()

        simulation_data = self._format_output(trajectories)
        if sim_status == SimulationReturnCode.PAUSED:
            simulation_data = self._make_resume_data(time_stopped, simulation_data, t)
        if resume is not None:
            simulation_data = self._update_resume_data(resume, simulation_data, time_stopped)
        self.result = simulation_data
        self.rc = int(sim_status)

        return sum([Results.build_from_solver_results(self, live_output_options)] * number_of_trajectories)


# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import tempfile

def cleanup_tempfiles():
    '''
    Cleanup all tempfiles in gillespy2 build.
    '''
    tempdir = tempfile.gettempdir()
    for file_obj in os.listdir(tempdir):
        if file_obj.startswith("gillespy2_build"):
            shutil.rmtree(os.path.join(tempdir, file_obj))



# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from datetime import time
import unittest
import numpy as np
import io
import gillespy2
from gillespy2.solvers import ODESolver, ODECSolver, NumPySSASolver, SSACSolver, TauLeapingCSolver, TauLeapingSolver, TauHybridSolver, TauHybridCSolver
from example_models import create_degradation

class TestRunOutput(unittest.TestCase):

    def setUp(self):
        self.model = create_degradation()
        tspan = gillespy2.TimeSpan.linspace(t=300, num_points=301)
        self.model.timespan(tspan)
        self.solvers = [
            # Solver and List of trajectory numbers to try
            (ODECSolver(model=self.model), [1]),
            (SSACSolver(model=self.model), [1, 3]),
            (TauLeapingCSolver(model=self.model), [1, 3]),
            (TauHybridCSolver(model=self.model), [1, 3]),
            (ODESolver(model=self.model), [1]),
            (NumPySSASolver(model=self.model), [1, 3]),
            (TauLeapingSolver(model=self.model), [1, 3]),
            (TauHybridSolver(model=self.model), [1, 3]),
        ]

    def test_run_output(self):
        expected_tspan = self.model.tspan
        expected_init = self.model.get_species("A").initial_value
        for solver, trajectory_counts in self.solvers:
            for number_of_trajectories in trajectory_counts:
                with self.subTest("Processing simulation output for each solver with different trajectory sizes",
                                  number_of_trajectories=number_of_trajectories,
                                  solver=solver):
                    results = self.model.run(solver=solver, number_of_trajectories=number_of_trajectories, seed=1024)
                    for n,result in enumerate(results):
                        tspan, first_value, last_value = result["time"], result["A"][0], result["A"][-1]
                        result_diff = np.concatenate([np.array([first_value]), result["A"][1:]])
                        result_diff = result["A"] - result_diff
                        self.assertTrue(np.allclose(tspan, expected_tspan), msg="Simulation output contains unexpected timeline values"
                                                                        f"\n  Received timeline: {tspan}\n  Expected timeline: {expected_tspan}, trajectory_number={n}")
                        self.assertEqual(first_value, expected_init, msg=f"Simulation output begins with unexpected value: {first_value}, trajectory_number={n}")
                        self.assertAlmostEqual(last_value, 0, places=3, msg=f"Simulation output converges on an unexpectedly large value: {last_value}, trajectory_number={n}")




# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Model exceptions
class ModelError(Exception):
    pass


class SpeciesError(ModelError):
    pass


class ParameterError(ModelError):
    pass


class ReactionError(ModelError):
    pass


class RateRuleError(ModelError):
    pass


class AssignmentRuleError(ModelError):
    pass


class EventError(ModelError):
    pass


class FunctionDefinitionError(ModelError):
    pass


class TimespanError(ModelError):
    pass


# Solver specific errors
class SolverError(Exception):
    pass


class DirectoryError(SolverError):
    pass


class BuildError(SolverError):
    pass


class ExecutionError(SolverError):
    pass


class SimulationError(Exception):
    pass


class StochMLImportError(SimulationError):
    pass


class InvalidStochMLError(SimulationError):
    pass


class InvalidModelError(SimulationError):
    pass


class SimulationTimeoutError(SimulationError):
    pass


# Results errors
class ResultsError(Exception):
    pass


class ValidationError(ResultsError):
    pass


class SBMLError(Exception):
    pass


# GillesPy2 is a modeling toolkit for biochemical simulation.
# Copyright (C) 2019-2024 GillesPy2 developers.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import copy

from .timespan import TimeSpan
from .gillespyError import SimulationError, ModelError

class GillesPySolver:
    """
    Abstract class for a solver.
    """

    name = "GillesPySolver"

    def run(self, model, t=20, number_of_trajectories=1, increment=0.05, seed=None,
            debug=False, profile=False, **kwargs):
        """
        Call out and run the solver. Collect the results.

        :param model: The model on which the solver will operate.
        :type model: gillespy.Model

        :param t: The end time of the solver
        :type t: int

        :param number_of_trajectories: The number of times to sample the chemical master equation. Each
            trajectory will be returned at the end of the simulation.
        :type number_of_trajectories: int

        :param increment: The time step of the solution
        :type increment: float

        :param seed: The random seed for the simulation. Defaults to None.
        :type seed: int

        :param debug: Set to True to provide additional debug information about the simulation.
        :type debug: bool

        :returns: Simulation trajectories.
        """
        raise SimulationError("This abstract solver class cannot be used directly.")

    def get_solver_settings(self):
        '''
        Abstract function for getting a solvers settings.
        '''
        raise SimulationError("This abstract solver class cannot be used directly")

    def validate_tspan(self, increment, t):
        """
        Validate the models time span and set it if not provided.

        :param increment: The current value of increment.
        :type increment: int

        :param t: The end time of the simulation.
        :type t: int

        :raises SimulationError: if timespan and increment are both set by the user or neither are set by the user.
        """
        if self.model.tspan is None and increment is None:
            raise SimulationError(
                """
                Failed while preparing to run the model. Neither increment or timespan are set.

                To continue either add a `timespan` definition to your Model or add the
                `increment` and `t` arguments to this `solver.run()` call.
                """
            )

        if self.model.tspan is not None and increment is not None:
            raise  SimulationError(
                """
                Failed while preparing to run the model. Both increment and timespan are set.

                To continue either remove your `timespan` definition from your Model or remove the
                `increment` argument from this `solver.run()` call.
                """
            )

        if self.model.tspan is None:
            if t is None:
                tspan = TimeSpan.arange(increment)
            else:
                tspan = TimeSpan.arange(increment, t=t)
            self.model.timespan(tspan)

    @classmethod
    def get_supported_features(cls) -> "Set[Type]":
        '''
        Get the set of supported features.
        '''
        return set()

    @classmethod
    def get_supported_integrator_options(cls) -> "Set[str]":
        '''
        Get the supported integrator options
        '''
        return set()

    @classmethod
    def validate_model(cls, sol_model, model):
        '''
        Validate that the solvers model is the same as the model passed to solver.run.

        :param sol_model: Solver model object.
        :type sol_model: gillespy2.Model

        :param model: Model object passed to solver.run.
        :type model: gillespy2.Model

        :raises SimulationError: If the models are not equal.
        '''
        if model is not None:
            model.compile_prep()
            if model.tspan is None:
                model = copy.deepcopy(model)
                model.tspan = sol_model.tspan
            if model.get_json_hash() != sol_model.get_json_hash():
                raise SimulationError("Model must equal ODECSolver.model.")

    @classmethod
    def validate_sbml_features(cls, model):
        '''
        Validate the models SBML features.

        :param model: The model object to be checked.
        :type model: gillespy2.Model

        :raises ModelError: If the model contains unsupported SBML features.
        '''
        unsupported_features = model.get_model_features() - cls.get_supported_features()
        if unsupported_features:
            unsupported_features = [feature.__name__ for feature in unsupported_features]
            raise ModelError(f"Could not run Model, "
                             f"SBML Features not supported by {cls.name}: " +
                             ", ".join(unsupported_features))

    @classmethod
    def validate_integrator_options(cls, options: "dict[str, float]") -> "dict[str, float]":
        '''
        Validate the integrator options.

        :param options: Integrator options to validate.
        :type options: dict

        :returns: Dict of supported integrator options.
        :rtype: dict
        '''
        return {
            option: value for option, value in options.items() if option in cls.get_supported_integrator_options()
        }
